@@ -137,7 +137,7 @@ void evi_remove_expired_subs(event_id_t id) {
 	while (subs) {
 		if (!subs->reply_sock) {
 			LM_ERR("unknown destination\n");
-			continue;
+			goto next_sub;
 		}
 		/* check expire */
 		if (!(subs->reply_sock->flags & EVI_PENDING) &&
@@ -163,6 +163,8 @@ void evi_remove_expired_subs(event_id_t id) {
 			}
 			continue;
 		}
+next_sub:
+		prev = subs;
 		subs = subs->next;
 	}
 	lock_release(events[id].lock);
@@ -769,6 +771,9 @@ mi_response_t *w_mi_subscribers_list_1(const mi_params_t *params,
 		return init_mi_error(404, MI_SSTR("Event not published"));
 	/* get the event id & before printing the subs list check for any expired subscribers and remove them*/
 	evid = evi_get_id(&event_s);
+	if (evid == -1)
+		return init_mi_error(404, MI_SSTR("Can't get the id"));
+
 	evi_remove_expired_subs(evid);
 
 	return mi_subscribers_list(event, NULL);
@@ -790,6 +795,9 @@ mi_response_t *w_mi_subscribers_list_2(const mi_params_t *params,
 		return init_mi_error(404, MI_SSTR("Event not published"));
 	/* get the event id & before printing the subs list check for any expired subscribers and remove them*/
 	evid = evi_get_id(&event_s);
+	if (evid == -1)
+		return init_mi_error(404, MI_SSTR("Can't get the id"));
+
 	evi_remove_expired_subs(evid);
 
 	if (get_mi_string_param(params, "socket", &subs_s.s, &subs_s.len) < 0)
