@@ -76,6 +76,7 @@
 static int fixup_check_pv_setf(void **param);
 static int fixup_time_rec(void **param);
 static int fixup_free_time_rec(void **param);
+static int fixup_spec_as_avp(void **param);
 
 static int set_prob(struct sip_msg *bar, int *percent_par);
 static int reset_prob(struct sip_msg*);
@@ -91,6 +92,7 @@ static int get_accurate_time(struct sip_msg* msg,
 static int pv_set_count(struct sip_msg* msg,
 					pv_spec_t *pv_name, pv_spec_t *pv_result);
 static int pv_sel_weight(struct sip_msg* msg, pv_spec_t *pv_name);
+static int w_shuffle_avps(struct sip_msg* msg, pv_spec_t *pv_name);
 
 mi_response_t *mi_set_prob(const mi_params_t *params,
 								struct mi_handler *async_hdl);
@@ -210,6 +212,10 @@ static const cmd_export_t cmds[]={
 		{CMD_PARAM_STR, 0, 0}, {0,0,0}},
 		REQUEST_ROUTE|FAILURE_ROUTE|ONREPLY_ROUTE|BRANCH_ROUTE|LOCAL_ROUTE|
 		STARTUP_ROUTE|TIMER_ROUTE|EVENT_ROUTE},
+	{"shuffle_avps",   (cmd_function)w_shuffle_avps,  {
+		{CMD_PARAM_VAR, fixup_spec_as_avp, NULL}, {0, 0, 0}},
+		ALL_ROUTES},
+
 	{0,0,{{0,0,0}},0}
 };
 
@@ -268,13 +274,13 @@ static const mi_export_t mi_cmds[] = {
 };
 
 static const pv_export_t mod_items[] = {
-	{ {"RANDOM", sizeof("RANDOM")-1}, 1000, pv_get_random_val, 0,
+	{ str_const_init("RANDOM"), 1000, pv_get_random_val, 0,
 		0, 0, 0, 0 },
-	{ {"shv", (sizeof("shv")-1)}, 1001, pv_get_shvar,
+	{ str_const_init("shv"), 1001, pv_get_shvar,
 		pv_set_shvar, pv_parse_shvar_name, 0, 0, 0},
-	{ {"ctime", (sizeof("ctime")-1)}, 1002, pv_get_time,
+	{ str_const_init("ctime"), 1002, pv_get_time,
 		0, pv_parse_time_name, 0, 0, 0},
-	{ {"env", (sizeof("env")-1)}, 1002, pv_get_env,
+	{ str_const_init("env"), 1002, pv_get_env,
 		0, pv_parse_env_name, 0, 0, 0},
 
 	{ {0, 0}, 0, 0, 0, 0, 0, 0, 0 }
@@ -338,6 +344,17 @@ static int fixup_free_time_rec(void **param)
 	return 0;
 }
 
+static int fixup_spec_as_avp(void **param)
+{
+	pv_spec_t *sp = (pv_spec_t*)(*param);
+	if (sp->type != PVT_AVP) {
+		LM_ERR("param spec must be an AVP\n");
+		return E_SCRIPT;
+	}
+
+	return 0;
+}
+
 /************************** module functions **********************************/
 
 mi_response_t *mi_set_prob(const mi_params_t *params,
@@ -699,7 +716,8 @@ static int get_accurate_time(struct sip_msg* msg,
 
 		val.flags = PV_VAL_STR;
 		val.rs.s = sec_usec_buf;
-		val.rs.len = sprintf(sec_usec_buf, "%ld.%06ld", tv.tv_sec, tv.tv_usec);
+		val.rs.len = sprintf(sec_usec_buf, "%lld.%06lld",
+						(long long)tv.tv_sec, (long long)tv.tv_usec);
 		if (pv_set_value(msg, pv_sec_usec, 0, &val) != 0) {
 			LM_ERR("failed to set 'pv_sec_usec'\n");
 			return -1;
@@ -871,3 +889,48 @@ int check_time_rec(struct sip_msg *_, char *time_rec, unsigned int *ptime)
 {
 	return _tmrec_expr_check_str(time_rec, ptime ? *ptime : time(NULL));
 }
+
+static int w_shuffle_avps(struct sip_msg* msg, pv_spec_t *pv_name)
+{
+	struct usr_avp *src_avp, *rnd_avp;
+	struct usr_avp *temp_avp = NULL;
+	int_str src_val, rnd_val;
+	unsigned short avp_type;
+	int avp_name;
+	int n, rnd_idx;
+
+	/* get the name */
+	if(pv_get_avp_name(msg, &pv_name->pvp, &avp_name, &avp_type)!=0)
+	{
+		LM_ERR("invalid name\n");
+		return -1;
+	}
+
+	/* count AVPs */
+	n = 0;
+	while ((temp_avp=search_first_avp(avp_type, avp_name, NULL, temp_avp)) != 0)
+		n++;
+
+	/* randomize AVPs */
+	for ( ; n>1; n-- ) {
+		rnd_idx = random() % n;
+		if (rnd_idx == (n-1))
+			continue;
+
+		LM_DBG("swapping [%d] <--> [%d]\n", (n-1), rnd_idx);
+
+		src_avp = search_index_avp(avp_type, avp_name, &src_val, (n-1));
+		rnd_avp = search_index_avp(avp_type, avp_name, &rnd_val, rnd_idx);
+
+		if ( replace_avp(avp_type|(rnd_avp->flags&AVP_VAL_STR), avp_name, rnd_val, (n-1))==-1 ||
+					replace_avp(avp_type|(src_avp->flags&AVP_VAL_STR), avp_name, src_val, rnd_idx)==-1 ) {
+			LM_ERR("failed to swap avp\n");
+			goto error;
+		}
+	}
+
+	return 1;
+error:
+	return -1;
+}
+
