@@ -487,19 +487,47 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 			case HDR_PPI_T:
 			case HDR_PAI_T:
 			case HDR_PRIVACY_T:
-			case HDR_RETRY_AFTER_T:
 			case HDR_CALL_INFO_T:
 			case HDR_WWW_AUTHENTICATE_T:
 			case HDR_PROXY_AUTHENTICATE_T:
+			case HDR_MIN_EXPIRES_T:
 			case HDR_FEATURE_CAPS_T:
+			case HDR_REPLACES_T:
+			case HDR_SECURITY_CLIENT_T:
+			case HDR_SECURITY_SERVER_T:
+			case HDR_SECURITY_VERIFY_T:
 				/* we ignore them for now even if they have something parsed*/
 				break;
 
+			/* not used in sip_msg, used in MSRP */
+			case HDR_TO_PATH_T:
+			case HDR_FROM_PATH_T:
+			case HDR_MESSAGE_ID_T:
+			case HDR_BYTE_RANGE_T:
+			case HDR_FAILURE_REPORT_T:
+			case HDR_SUCCESS_REPORT_T:
+			case HDR_STATUS_T:
+			case HDR_USE_PATH_T:
+
+			/* not having shortcut */
+			case HDR_RETRY_AFTER_T:
+			case HDR_VIA2_T:
+
+			/* not actual hdrs */
+			case HDR_OTHER_T:
+			case HDR_ERROR_T:
+			case HDR_EOH_T:
+				break;
+
+			/* we do not have a "default" on purpose, so we get
+			 * a compile err/war where a new HDR is added and we do 
+			 * not handle it here.
 			default:
 				if (hdr->parsed) {
 					LM_WARN("header body ignored: %d\n", hdr->type );
 				}
 				break;
+			*/
 		}/*switch*/
 	}/*for all headers*/
 
@@ -636,7 +664,6 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 				else if ( !new_msg->via2 )
 				{
 					LINK_SIBLING_HEADER(h_via1, new_hdr);
-					new_msg->h_via2 = new_hdr;
 					new_msg->via2 = via_body_cloner( new_msg->buf,
 						org_msg->buf, (struct via_body*)hdr->parsed, &p);
 					new_hdr->parsed  = (void*)new_msg->via2;
@@ -935,6 +962,39 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 					new_msg->proxy_authenticate = new_hdr;
 				}
 				break;
+			case HDR_MIN_EXPIRES_T:
+				if (HOOK_NOT_SET(min_expires)) {
+					new_msg->min_expires = new_hdr;
+				}
+				break;
+			case HDR_FEATURE_CAPS_T:
+				if (HOOK_NOT_SET(feature_caps)) {
+					new_msg->feature_caps = new_hdr;
+				}
+				break;
+			case HDR_REPLACES_T:
+				if (HOOK_NOT_SET(replaces)) {
+					new_msg->replaces = new_hdr;
+				}
+				break;
+			case HDR_SECURITY_CLIENT_T:
+				if (HOOK_NOT_SET(security_client))
+					new_msg->security_client = new_hdr;
+				else
+					LINK_SIBLING_HEADER(security_client, new_hdr);
+				break;
+			case HDR_SECURITY_SERVER_T:
+				if (HOOK_NOT_SET(security_server))
+					new_msg->security_server = new_hdr;
+				else
+					LINK_SIBLING_HEADER(security_server, new_hdr);
+				break;
+			case HDR_SECURITY_VERIFY_T:
+				if (HOOK_NOT_SET(security_verify))
+					new_msg->security_verify = new_hdr;
+				else
+					LINK_SIBLING_HEADER(security_verify, new_hdr);
+				break;
 			default:
 				/* ignore the rest*/
 				;
