@@ -35,6 +35,7 @@
 #include "../../parser/parse_authenticate.h"
 #include "../../parser/contact/parse_contact.h"
 #include "../../parser/parse_min_expires.h"
+#include "../../parser/parse_expires.h"
 #include "../uac_auth/uac_auth.h"
 #include "../../lib/digest_auth/digest_auth.h"
 #include "reg_records.h"
@@ -98,6 +99,8 @@ static mi_response_t *mi_reg_enable(const mi_params_t *params,
 								struct mi_handler *async_hdl);
 static mi_response_t *mi_reg_disable(const mi_params_t *params,
 								struct mi_handler *async_hdl);
+static mi_response_t *mi_reg_force_register(const mi_params_t *params,
+								struct mi_handler *async_hdl);
 
 int send_register(unsigned int hash_index, reg_record_t *rec, str *auth_hdr);
 int send_unregister(unsigned int hash_index, reg_record_t *rec, str *auth_hdr,
@@ -110,6 +113,7 @@ uac_auth_api_t uac_auth_api;
 
 unsigned int default_expires = 3600;
 unsigned int timer_interval = 100;
+unsigned int failure_retry_interval = 0;
 
 reg_table_t reg_htable = NULL;
 unsigned int reg_hsize = 1;
@@ -139,6 +143,7 @@ static const param_export_t params[]= {
 	{"hash_size",		INT_PARAM,			&reg_hsize},
 	{"default_expires",	INT_PARAM,			&default_expires},
 	{"timer_interval",	INT_PARAM,			&timer_interval},
+	{"failure_retry_interval", INT_PARAM,			&failure_retry_interval},
 	{"enable_clustering",	INT_PARAM,			&enable_clustering},
 	{"db_url",		STR_PARAM,			&db_url.s},
 	{"table_name",		STR_PARAM,			&reg_table_name.s},
@@ -178,6 +183,10 @@ static const mi_export_t mi_cmds[] = {
 		{mi_reg_disable, {"aor", "contact", "registrar", 0}},
 		{EMPTY_MI_RECIPE}}
 	},
+	{ "reg_force_register", 0, 0, 0, {
+		{mi_reg_force_register, {"aor", "contact", "registrar", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
 	{EMPTY_MI_EXPORT}
 };
 
@@ -354,6 +363,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 	str msg_body;
 	int statuscode = 0;
 	unsigned int exp = 0;
+	exp_body_t *header_exp = NULL;
 	unsigned int bindings_counter = 0;
 	reg_record_t *rec = (reg_record_t*)e_data;
 	struct hdr_field *c_ptr, *head_contact;
@@ -430,7 +440,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 						rec->last_register_sent = now;
 						rec->state = REGISTERING_STATE;
 					} else {
-						rec->registration_timeout = now + rec->expires - timer_interval;
+						rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 						rec->state = INTERNAL_ERROR_STATE;
 					}
 				}
@@ -481,8 +491,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 		contact = ((contact_body_t*)msg->contact->parsed)->contacts;
 		while (contact) {
 			/* Check for binding */
-			if (contact->uri.len==rec->contact_uri.len &&
-				strncmp(contact->uri.s,rec->contact_uri.s,contact->uri.len)==0){
+			if (compare_uris(&contact->uri, NULL, &rec->contact_uri, NULL) == 0){
 				if (contact->expires && contact->expires->body.len) {
 					if (str2int(&contact->expires->body, &exp)<0) {
 						LM_ERR("Unable to extract expires from [%.*s]"
@@ -511,6 +520,15 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 				contact = contact->next;
 			}
 		}
+		// if there is no expires in contact, try parse expires from header
+		if (exp == 0 && msg->expires) {
+			if (parse_expires(msg->expires) < 0) {
+				LM_ERR("failed to parse Expires header\n");
+			} else {
+				header_exp = (exp_body_t*)msg->expires->parsed;
+				exp = header_exp->val;
+			}
+		}
 
 		if (!(rec->flags&REG_ENABLED)) {
 			if (rec->state==REGISTERING_STATE ||
@@ -541,7 +559,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 					rec->last_register_sent = now;
 					rec->state = REGISTERING_STATE;
 				} else {
-					rec->registration_timeout = now + rec->expires - timer_interval;
+					rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 					rec->state = INTERNAL_ERROR_STATE;
 				}
 			} else {
@@ -572,7 +590,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 			rec->auth_password.s==NULL || rec->auth_password.len==0) {
 			LM_ERR("Credentials not provisioned\n");
 			rec->state = WRONG_CREDENTIALS_STATE;
-			rec->registration_timeout = 0;
+			rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 			/* action successfully completed on current list element */
 			return 1; /* exit list traversal */
 		}
@@ -607,7 +625,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 			LM_WARN("Wrong credentials for [%.*s]\n",
 				rec->td.rem_uri.len, rec->td.rem_uri.s);
 			rec->state = WRONG_CREDENTIALS_STATE;
-			rec->registration_timeout = 0;
+			rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 			/* action successfully completed on current list element */
 			return 1; /* exit list traversal */
 		default:
@@ -681,13 +699,13 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 				rec->state = INTERNAL_ERROR_STATE;
 		} else {
 			rec->state = REGISTRAR_ERROR_STATE;
-			rec->registration_timeout = now + rec->expires - timer_interval;
+			rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 		}
 		break;
 
 	case 408: /* Interval Too Brief */
 		rec->state = REGISTER_TIMEOUT_STATE;
-		rec->registration_timeout = now + rec->expires - timer_interval;
+		rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 		break;
 
 	default:
@@ -697,7 +715,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 		} else {
 			/* we got an error from the server */
 			rec->state = REGISTRAR_ERROR_STATE;
-			rec->registration_timeout = now + rec->expires - timer_interval;
+			rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 
 		}
 	}
@@ -706,7 +724,7 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 	return 1; /* exit list traversal */
 done:
 	rec->state = INTERNAL_ERROR_STATE;
-	rec->registration_timeout = now + rec->expires;
+	rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 	return -1; /* exit list traversal */
 }
 
@@ -936,13 +954,18 @@ int run_timer_check(void *e_data, void *data, void *r_data)
 	case INTERNAL_ERROR_STATE:
 	case REGISTRAR_ERROR_STATE:
 		reg_print_record(rec);
+		/* don't retry failed registrations too soon */
+		if (now < rec->registration_timeout) {
+			LM_DBG("won't retry registration for another %lld secs\n", (long long)(rec->registration_timeout - now));
+			break;
+		}
 		if (rec->flags&REG_ENABLED) {
 			new_call_id_ftag_4_record(rec, s_now);
 			if(send_register(i, rec, NULL)==1) {
 				rec->last_register_sent = now;
 				rec->state = REGISTERING_STATE;
 			} else {
-				rec->registration_timeout = now + rec->expires - timer_interval;
+				rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 				rec->state = INTERNAL_ERROR_STATE;
 			}
 		} else {
@@ -964,7 +987,7 @@ int run_timer_check(void *e_data, void *data, void *r_data)
 				rec->last_register_sent = now;
 				rec->state = REGISTERING_STATE;
 			} else {
-				rec->registration_timeout = now + rec->expires - timer_interval;
+				rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 				rec->state = INTERNAL_ERROR_STATE;
 			}
 		}
@@ -1125,6 +1148,14 @@ int run_mi_reg_list(void *e_data, void *data, void *r_data)
 			rec->td.loc_uri.s, rec->td.loc_uri.len) < 0)
 			goto error;
 
+	if (rec->td.send_sock)
+	{
+		p = socket2str(rec->td.send_sock, NULL, &len, 0);
+		if (p)
+			if (add_mi_string(record_item, MI_SSTR("socket"), p, len) < 0)
+				goto error;
+	}
+
 	if (rec->td.obp.s && rec->td.obp.len)
 		if (add_mi_string(record_item, MI_SSTR("proxy"),
 			rec->td.obp.s, rec->td.obp.len) < 0)
@@ -1395,7 +1426,7 @@ int run_mi_reg_enable(void *e_data, void *data, void *r_data)
 					rec->last_register_sent = now;
 					rec->state = REGISTERING_STATE;
 				} else {
-					rec->registration_timeout = now + rec->expires - timer_interval;
+					rec->registration_timeout = now + (failure_retry_interval?failure_retry_interval:rec->expires) - timer_interval;
 					rec->state = INTERNAL_ERROR_STATE;
 				}
 			}
@@ -1488,3 +1519,48 @@ static mi_response_t *mi_reg_disable(const mi_params_t *params,
 
 	return init_mi_result_ok();
 }
+
+
+int run_mi_reg_force_register(void *e_data, void *data, void *r_data)
+{
+	reg_record_t *rec = (reg_record_t*)e_data;
+	record_coords_t *coords = (record_coords_t *)data;
+
+	if (!str_strcmp(&coords->contact, &rec->contact_uri) &&
+		!str_strcmp(&coords->registrar, &rec->td.rem_target)) {
+		if (rec->flags&REG_ENABLED) {
+			rec->registration_timeout = 0;
+		}
+
+		return 1;
+	} else
+		return 0;  /* continue search */
+}
+
+static mi_response_t *mi_reg_force_register(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	mi_response_t *resp;
+	record_coords_t coords;
+	int rc;
+	unsigned int hash_code;
+
+	if ((resp = mi_get_coords(params, &coords)))
+		return resp;
+
+	hash_code = core_hash(&coords.aor, NULL, reg_hsize);
+	coords.extra = (void*)(unsigned long)hash_code;
+
+	lock_get(&reg_htable[hash_code].lock);
+	rc = slinkedl_traverse(reg_htable[hash_code].p_list,
+					&run_mi_reg_force_register, &coords, NULL);
+	lock_release(&reg_htable[hash_code].lock);
+
+	if (rc < 0)
+		return NULL;
+	else if (rc == 0)
+		return init_mi_error(404, MI_SSTR("No such registrant"));
+
+	return init_mi_result_ok();
+}
+
